@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1997-2018, University of Amsterdam
+    Copyright (c)  1997-2019, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -538,9 +539,9 @@ prolog_message(commandline_arg_type(Flag, Arg)) -->
     [ 'Bad argument to commandline option -~w: ~w'-[Flag, Arg] ].
 prolog_message(missing_feature(Name)) -->
     [ 'This version of SWI-Prolog does not support ~w'-[Name] ].
-prolog_message(singletons(List)) -->
+prolog_message(singletons(_Term, List)) -->
     [ 'Singleton variables: ~w'-[List] ].
-prolog_message(multitons(List)) -->
+prolog_message(multitons(_Term, List)) -->
     [ 'Singleton-marked variables appearing more than once: ~w'-[List] ].
 prolog_message(profile_no_cpu_time) -->
     [ 'No CPU-time info.  Check the SWI-Prolog manual for details' ].
@@ -615,10 +616,12 @@ prolog_message(no_exported_op(Module, Op)) -->
 prolog_message(discontiguous((-)/2,_)) -->
     prolog_message(minus_in_identifier).
 prolog_message(discontiguous(Proc,Current)) -->
-    [ 'Clauses of ~p are not together in the source-file'-[Proc], nl ],
-    current_definition(Proc, '  Earlier definition at '),
-    [ '  Current predicate: ~p'-[Current], nl,
-      '  Use :- discontiguous ~p. to suppress this message'-[Proc]
+    [ 'Clauses of ', ansi(code, '~p', [Proc]),
+      ' are not together in the source-file', nl ],
+    current_definition(Proc, 'Earlier definition at '),
+    [ 'Current predicate: ', ansi(code, '~p', [Current]), nl,
+      'Use ', ansi(code, ':- discontiguous ~p.', [Proc]),
+      ' to suppress this message'
     ].
 prolog_message(decl_no_effect(Goal)) -->
     [ 'Deprecated declaration has no effect: ~p'-[Goal] ].
@@ -762,26 +765,6 @@ prolog_message(file_search(tried(Spec, Cond), Path)) -->
                  *              GC              *
                  *******************************/
 
-prolog_message(gc(start)) -->
-    thread_context,
-    [ 'GC: ', flush ].
-prolog_message(gc(done(G, T, Time, UG, UT, RG, RT))) -->
-    [ at_same_line,
-      'gained ~D+~D in ~3f sec; used ~D+~D; free ~D+~D'-
-      [G, T, Time, UG, UT, RG, RT]
-    ].
-prolog_message(shift_stacks(start(_L,_G,_T))) -->
-    thread_context,
-    [ 'Stack-shift: ', flush ].
-prolog_message(shift_stacks(done(Time, L, G, T))) -->
-    { LKB is L//1024,
-      GKB is G//1024,
-      TKB is T//1024
-    },
-    [ at_same_line,
-      'local: ~DKB, global: ~DKB, trail: ~DKB bytes (~2f sec)'-
-      [LKB, GKB, TKB, Time]
-    ].
 prolog_message(agc(start)) -->
     thread_context,
     [ 'AGC: ', flush ].
@@ -1041,14 +1024,17 @@ prolog_message(copyright) -->
       'Please run ?- license. for legal details.'
     ].
 prolog_message(user_versions) -->
-    { findall(Msg, prolog:version_msg(Msg), Msgs) },
-    user_version_messages(Msgs).
+    (   { findall(Msg, prolog:version_msg(Msg), Msgs),
+          Msgs \== []
+        }
+    ->  [nl],
+        user_version_messages(Msgs)
+    ;   []
+    ).
 prolog_message(documentaton) -->
     [ 'For online help and background, visit http://www.swi-prolog.org', nl,
       'For built-in help, use ?- help(Topic). or ?- apropos(Word).'
     ].
-prolog_message(author) -->
-    [ 'Jan Wielemaker (jan@swi-prolog.org)' ].
 prolog_message(welcome) -->
     [ 'Welcome to SWI-Prolog (' ],
     prolog_message(threads),
@@ -1063,11 +1049,12 @@ prolog_message(welcome) -->
     prolog_message(documentaton),
     [ nl, nl ].
 prolog_message(about) -->
-    [ 'SWI-Prolog version ' ],
+    [ 'SWI-Prolog version (' ],
+    prolog_message(threads),
+    prolog_message(address_bits),
+    ['version ' ],
     prolog_message(version),
-    [ ' by ' ],
-    prolog_message(author),
-    [ nl ],
+    [ ')', nl ],
     prolog_message(copyright).
 prolog_message(halt) -->
     [ 'halt' ].
@@ -1466,9 +1453,8 @@ prolog_message(pack(no_arch(Entry, Arch))) -->
 
 prolog_message(null_byte_in_path(Component)) -->
     [ '0-byte in PATH component: ~p (skipped directory)'-[Component] ].
-prolog_message(invalid_tmp_var(Var, Value, Reason)) -->
-    [ 'Cannot use '-[] ], env(Var),
-    [ ' as temporary file directory: ~p: ~w'-[Value, Reason] ].
+prolog_message(invalid_tmp_dir(Dir, Reason)) -->
+    [ 'Cannot use ~p as temporary file directory: ~w'-[Dir, Reason] ].
 prolog_message(ambiguous_stream_pair(Pair)) -->
     [ 'Ambiguous operation on stream pair ~p'-[Pair] ].
 
@@ -1542,7 +1528,7 @@ print_system_message(Term, Kind, Lines) :-
     Term \= error(syntax_error(_), _),
     msg_property(Kind, location_prefix(File:Line, LocPrefix, LinePrefix)),
     !,
-    insert_prefix(Lines, LinePrefix, PrefixLines),
+    insert_prefix(Lines, LinePrefix, Ctx, PrefixLines),
     '$append'([ begin(Kind, Ctx),
                 LocPrefix,
                 nl
@@ -1578,10 +1564,12 @@ msg_property(query, stream(user_output)) :- !.
 msg_property(_, stream(user_error)) :- !.
 msg_property(error,
              location_prefix(File:Line,
-                             '~NERROR: ~w:~d:'-[File,Line], '~N\t')) :- !.
+                             '~NERROR: ~w:~d:'-[File,Line],
+                             '~NERROR:    ')) :- !.
 msg_property(warning,
              location_prefix(File:Line,
-                             '~NWarning: ~w:~d:'-[File,Line], '~N\t')) :- !.
+                             '~NWarning: ~w:~d:'-[File,Line],
+                             '~NWarning:    ')) :- !.
 msg_property(error,   wait(0.1)) :- !.
 
 msg_prefix(debug(_), Prefix) :-
@@ -1649,7 +1637,7 @@ add_message_context1(thread, Prefix0, Prefix) :-
 print_message_lines(Stream, kind(Kind), Lines) :-
     !,
     msg_property(Kind, prefix(Prefix)),
-    insert_prefix(Lines, Prefix, PrefixLines),
+    insert_prefix(Lines, Prefix, Ctx, PrefixLines),
     '$append'([ begin(Kind, Ctx)
               | PrefixLines
               ],
@@ -1658,25 +1646,29 @@ print_message_lines(Stream, kind(Kind), Lines) :-
               AllLines),
     print_message_lines(Stream, AllLines).
 print_message_lines(Stream, Prefix, Lines) :-
-    insert_prefix(Lines, Prefix, PrefixLines),
+    insert_prefix(Lines, Prefix, _, PrefixLines),
     print_message_lines(Stream, PrefixLines).
 
-%!  insert_prefix(+Lines, +Prefix, -PrefixedLines)
+%!  insert_prefix(+Lines, +Prefix, +Ctx, -PrefixedLines)
 
-insert_prefix([at_same_line|Lines0], Prefix, Lines) :-
+insert_prefix([at_same_line|Lines0], Prefix, Ctx, Lines) :-
     !,
-    prefix_nl(Lines0, Prefix, Lines).
-insert_prefix(Lines0, Prefix, [prefix(Prefix)|Lines]) :-
-    prefix_nl(Lines0, Prefix, Lines).
+    prefix_nl(Lines0, Prefix, Ctx, Lines).
+insert_prefix(Lines0, Prefix, Ctx, [prefix(Prefix)|Lines]) :-
+    prefix_nl(Lines0, Prefix, Ctx, Lines).
 
-prefix_nl([], _, [nl]).
-prefix_nl([nl], _, [nl]) :- !.
-prefix_nl([flush], _, [flush]) :- !.
-prefix_nl([nl|T0], Prefix, [nl, prefix(Prefix)|T]) :-
+prefix_nl([], _, _, [nl]).
+prefix_nl([nl], _, _, [nl]) :- !.
+prefix_nl([flush], _, _, [flush]) :- !.
+prefix_nl([nl|T0], Prefix, Ctx, [nl, prefix(Prefix)|T]) :-
     !,
-    prefix_nl(T0, Prefix, T).
-prefix_nl([H|T0], Prefix, [H|T]) :-
-    prefix_nl(T0, Prefix, T).
+    prefix_nl(T0, Prefix, Ctx, T).
+prefix_nl([ansi(Attrs,Fmt,Args)|T0], Prefix, Ctx,
+          [ansi(Attrs,Fmt,Args,Ctx)|T]) :-
+    !,
+    prefix_nl(T0, Prefix, Ctx, T).
+prefix_nl([H|T0], Prefix, Ctx, [H|T]) :-
+    prefix_nl(T0, Prefix, Ctx, T).
 
 %!  print_message_lines(+Stream, +Lines)
 
@@ -1701,24 +1693,46 @@ line_element(S, nl) :-
     nl(S).
 line_element(S, prefix(Fmt-Args)) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
 line_element(S, prefix(Fmt)) :-
     !,
-    format(S, Fmt, []).
+    safe_format(S, Fmt, []).
 line_element(S, flush) :-
     !,
     flush_output(S).
 line_element(S, Fmt-Args) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
 line_element(S, ansi(_, Fmt, Args)) :-
     !,
-    format(S, Fmt, Args).
+    safe_format(S, Fmt, Args).
+line_element(S, ansi(_, Fmt, Args, _Ctx)) :-
+    !,
+    safe_format(S, Fmt, Args).
 line_element(_, begin(_Level, _Ctx)) :- !.
 line_element(_, end(_Ctx)) :- !.
 line_element(S, Fmt) :-
-    format(S, Fmt, []).
+    safe_format(S, Fmt, []).
 
+%!  safe_format(+Stream, +Format, +Args) is det.
+
+safe_format(S, Fmt, Args) :-
+    E = error(_,_),
+    catch(format(S,Fmt,Args), E,
+          format_failed(S,Fmt,Args,E)).
+
+format_failed(S, _Fmt, _Args, E) :-
+    E = error(io_error(_,S),_),
+    !,
+    throw(E).
+format_failed(S, Fmt, Args, error(E,_)) :-
+    format(S, '~N    [[ EXCEPTION while printing message ~q~n\c
+                        ~7|with arguments ~W:~n\c
+                        ~7|raised: ~W~n~4|]]~n',
+           [ Fmt,
+             Args, [quoted(true), max_depth(10)],
+             E, [quoted(true), max_depth(10)]
+           ]).
 
 %!  message_to_string(+Term, -String)
 %
