@@ -253,6 +253,12 @@ add_global_worklist(worklist *wl)
 }
 
 
+/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Normal completion is done. There  may   be  worklists that are suspended
+using negation_suspend/3. We wake  these  up   by  adding  a  new answer
+cluster with a NULL node.
+- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
+
 static worklist *
 negative_worklist(tbl_component *c)
 { worklist **wlp = baseBuffer(&c->created_worklists->members, worklist*);
@@ -268,7 +274,7 @@ negative_worklist(tbl_component *c)
 
       wl->neg_complete = TRUE;
       DEBUG(MSG_TABLING_NEG,
-	    Sdprintf("Resume negative node %p\n", wl));
+	    Sdprintf("Resume negative node %zd\n", pointerToInt(wl)));
 
       c = new_answer_cluster(NULL);
       wkl_append_left(wl, c);
@@ -1631,7 +1637,8 @@ unify_cluster(term_t t, cluster *c, int is_riac)
     term_t modeav = PL_new_term_ref();
 
     for(; ap < top; ap++)
-    { if ( !PL_unify_list(tail, head, tail) ||
+    { if ( !PL_put_variable(answer) ||
+	   !PL_unify_list(tail, head, tail) ||
 	   !tbl_unify_trie_term(*ap, answer PASS_LD) ||
 	   !tbl_put_trie_value(modeav, *ap PASS_LD) ||
 	   !PL_unify_term(head, PL_FUNCTOR_CHARS, "-", 2,
